/* Copyright (c) 2004 Limor Fried <ladyada@media.mit.edu>, MIT Media Lab

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*/

#include <inttypes.h>
#include <avr/io.h>
#include <stdio.h>
#include "main.h"
#include "delay.h"
#include "cs8900a.h"


void cs8900aInit(void) {
  crystalEnable();

  if (!crystalVerify()) {
    printf("CS8900A not found!\n\r");
    halt();
  }

  crystalReset();

  crystalWriteData(TEST_CTL, TEST_CTL_FDX);        //set to full duplex
   
  crystalWriteData(RX_CTL, (/*RX_CTL_PROMISC |*/ RX_CTL_RX_OK_A |
			    RX_CTL_IND_A | RX_CTL_BCAST_A));
  crystalWriteData(IND_ADDR, (MYMAC_1<<8|MYMAC_0));      //  ;0xbbaa Write
  crystalWriteData(IND_ADDR+2, (MYMAC_3<<8|MYMAC_2));    //  ;0xddcc  out
  crystalWriteData(IND_ADDR+4, (MYMAC_5<<8|MYMAC_4));    //  ;0xffee   48 bit IA
  
  crystalWriteData(LINE_CTL, LINE_CTL_10BASET |
		   LINE_CTL_TX_ON | LINE_CTL_RX_ON);
  
  printf("CS8900a Initialized\n\r");
}

unsigned int cs8900aBeginPacketRetreive(void) {
  unsigned int packetlen = 0;

  // first poll to check for a packet waiting
  if (!crystalPoll()) {
     return 0;
  }
  // rad, there is a packet in the queue
  crystalIORead(RX_TX_DATA+1);      // read the rx status register
  crystalIORead(RX_TX_DATA); 
  
  packetlen = crystalIORead(RX_TX_DATA+1) << 8;     // read the frame length
  packetlen |= crystalIORead(RX_TX_DATA);

  // and return it
  return packetlen;
}

void cs8900aEndPacketRetreive() {
  crystalWriteData(RX_CTL, crystalReadData(RX_CTL) | RX_CFG_SKIP);
}

void cs8900aRetreivePacketData(unsigned char * localBuffer, unsigned int length)
{
  uint i;
  for (i=0; i<length;i+=2) {
    localBuffer[i] = crystalIORead(RX_TX_DATA);
    localBuffer[i+1] = crystalIORead(RX_TX_DATA+1);
  }
}


void cs8900aSendPacketData(unsigned char * localBuffer, unsigned int length)
{
  uint16_t i;

  for (i=0; i<length; i+=2) {
    crystalIOWrite(RX_TX_DATA, localBuffer[i]);
    crystalIOWrite(RX_TX_DATA+1, localBuffer[i+1]);
    //printf("%02x%02x ", localBuffer[i], localBuffer[i+1]);
  } 
}

void cs8900aEndPacketSend(void)
{
  
}

void cs8900aBeginPacketSend(uint framelen) {
  // allocate that # bytes!
  uint16_t ret;
  
  // start transmission
  crystalIOWrite(TX_CMD, LO( TX_CMD_START_ALL ));
  crystalIOWrite(TX_CMD+1, HI( TX_CMD_START_ALL ));

  crystalIOWrite(TX_LEN, LO(framelen));
  crystalIOWrite(TX_LEN+1, HI(framelen));

  ret = crystalReadData(BUS_STATUS);
  if (ret & TX_BID_ERR) {
    printf("Tx Bidding error\n\r");
    return;
  }
  while (! crystalReadData(BUS_STATUS) & BUS_ST_RDY4TXNOW) {}
  // bus ready
}
/*****************************************************************/

void crystalEnable(void) {
  cbi(CONTROLPORT, CRYSTAL_AEN);
  sbi(CONTROLPORT, CRYSTAL_IOR);
  sbi(CONTROLPORT, CRYSTAL_IOW);
  sbi(CONTROLPORT, CRYSTAL_RESET);
  delay_ms(1);
  cbi(CONTROLPORT, CRYSTAL_RESET);
  //printf("CS8900a Enabled\n\r");
  delay_ms(10);
}


byte crystalVerify() {
   byte datalo, datahi;
   uint16_t u;

   datalo = crystalIORead(PORTPTR);
   datahi = crystalIORead(PORTPTR+1);
   printf("Chip query returned 0x%x%x\n\r",datahi, datalo);

   if ((datahi & 0xF0) == 0x30) {
      // valid
      u = crystalReadData(EISA);
      printf("EISA = 0x%x ", u);
      if (u != 0x630e) {
         putstring("Problem with CS8900A\n\r");
         return(0);
      }
      u = crystalReadData(PRODID);
      printf("ProdID = 0x%x%x\n\r", HI(u), LO(u));
      return(1);
   } else {
	return(0);
   }
}


void crystalReset() {
   uint16_t d;

   crystalWriteData(SELF_CTL, SELF_CTL_RESET);
   while (1) {
      delay_ms(1);
      d = crystalReadData(SELF_CTL);
      //DEBUG(printf("self ctl = 0x%x%x\n\r", HI(d), LO(d));)
      if (d & 0x40) {
      	continue;
      }
      d = crystalReadData(SELF_STATUS);
      //DEBUG(printf("self status = 0x%x%x\n\r", HI(d), LO(d));)
	if (! d & 0x80) {
	  continue;
	}
      break;
   }
   //puts("CS8900 has been reset\n\r");
}



void crystalInit(byte promisc, byte *MAC) {
   crystalWriteData(TEST_CTL, TEST_CTL_FDX);        //set to full duplex

   crystalWriteData(RX_CTL, (/*RX_CTL_PROMISC |*/ RX_CTL_RX_OK_A |
			     RX_CTL_IND_A | RX_CTL_BCAST_A));
   crystalWriteData(IND_ADDR, (MAC[1]<<8|MAC[0]));      //  ;0xbbaa Write
   crystalWriteData(IND_ADDR+2, (MAC[3]<<8|MAC[2]));    //  ;0xddcc  out
   crystalWriteData(IND_ADDR+4, (MAC[5]<<8|MAC[4]));    //  ;0xffee   48 bit IA

   crystalWriteData(LINE_CTL, LINE_CTL_10BASET |
		    LINE_CTL_TX_ON | LINE_CTL_RX_ON);

   // printf("CS8900a Initialized\n\r");
}




// returns TRUE if theres something, FALSE if not
byte crystalPoll() {
   uint16_t d;

   d = crystalReadData(RX_EVT);
   //printf("rx evt = %x%x\n\r", HI(d), LO(d));
   if (d & 0x0100) {
      return 1;
   } else {
      return 0;
   }
}



/*************** LOW LEVEL STUFF ***************/


void crystalSetAddr(byte addr) {
  //printf("setting address to addr %x", addr);
  outb(ADDRPORT, addr & 0xf);
}

//#define crystalSetAddr(i) output_b( (input_b()&0xf0) | ( i & 0x0f))

void crystalSetPointer(uint16_t addr) {
  //printf("ptr = 0x%x%x\n\r", (addr >> 8) & 0xff, addr & 0xff);
   crystalIOWrite(PORTPTR, addr & 0xff);
   crystalIOWrite(PORTPTR+1, (addr >> 8) & 0xff);
}

//#define crystalSetPointer(a) crystalIOWrite(PORTPTR, a & 0xff); crystalIOWrite(PORTPTR+1, ( a >> 8) & 0xff)


void crystalIOWrite(byte addr, byte data) {
  crystalSetAddr(addr);
  outb(DATAPORTTRIS, 0xff);        // set to output
  outb(DATAPORT, data);

  cbi(CONTROLPORT, CRYSTAL_AEN);
  NOP; NOP; NOP;
  cbi(CONTROLPORT, CRYSTAL_IOW);
  NOP; NOP; NOP;
  sbi(CONTROLPORT, CRYSTAL_IOR);
  NOP; NOP; NOP; NOP;
  sbi(CONTROLPORT, CRYSTAL_IOW);
  NOP; NOP; NOP; NOP;
  sbi(CONTROLPORT, CRYSTAL_AEN);
}

byte crystalIORead(byte addr) {
   byte val;

   cbi(CONTROLPORT, CRYSTAL_AEN);
   NOP; NOP; NOP;
   crystalSetAddr(addr);

   //step();
   DATAPORTTRIS = 0x00;        // set to input
   NOP; NOP; NOP; NOP;
   sbi(CONTROLPORT, CRYSTAL_IOW);
   cbi(CONTROLPORT, CRYSTAL_IOR);
   NOP; NOP; NOP; NOP;
   //   puts("read ");
   //step();
   val = PINC;
   
   sbi(CONTROLPORT, CRYSTAL_IOR);
   sbi(CONTROLPORT, CRYSTAL_AEN);

   return val;
}

uint16_t crystalIORead16(byte addr) {
   uint16_t data;

   data = crystalIORead(addr+1) << 8;
   data |= crystalIORead(addr);

   return data;
}

uint16_t crystalReadData(uint16_t addr) {
  uint16_t val;

   crystalSetPointer(addr);
   val = crystalIORead(PORTDATA+1) << 8;             // high byte
   val |= crystalIORead(PORTDATA);              // low byte

   return val;
}

void crystalWriteData(uint16_t addr, uint16_t data) {
   crystalSetPointer(addr);
   crystalIOWrite(PORTDATA, data & 0xff);            // send low byte
   crystalIOWrite(PORTDATA+1, (data>>8) & 0xff);     // send high byte
}

//#define crystalWriteData(a, d) crystalSetPointer( a ); crystalIOWrite(PORTDATA, d & 0xff); (PORTDATA+1, ( d >>8) & 0xff)
