#include <inttypes.h>
#include <string.h>
#include <stdio.h>
#include "uipopt.h"
#include "udp.h"
#include "dhcp.h"
#include "main.h"
#include "uip_arp.h"
#include "uip.h"
#include "nic.h"
#include "dnslookup.h"

char hostname[] = "control0";
char vendorname[] = "aether";

uint16_t dhcp_xid1, dhcp_xid2;       // our unique identifier

dhcpStatus dhcp_status;
uint16_t dhcp_timeout;

#define UDPBUF ((uip_udpip_hdr *)&uip_buf[UIP_LLH_LEN])
#define ETHBUF ((struct uip_eth_hdr *)&uip_buf[0])


/*
 * DHCP IP Aquisition steps
 * -Send DHCPDISCOVER packet
 * -Receive DHCPOFFER packet
 * -Confirm offer with DHCPREQUEST packet
 * -Receive DHCPACK packet

                Server          Client          Server
            (not selected)                    (selected)

                  v               v               v
                  |               |               |
                  |     Begins initialization     |
                  |               |               |
                  | _____________/|\____________  |
                  |/DHCPDISCOVER | DHCPDISCOVER  \|
                  |               |               |
              Determines          |          Determines
             configuration        |         configuration
                  |               |               |
                  |\             |  ____________/ |
                  | \________    | /DHCPOFFER     |
                  | DHCPOFFER\   |/               |
                  |           \  |                |
                  |       Collects replies        |
                  |             \|                |
                  |     Selects configuration     |
                  |               |               |
                  | _____________/|\____________  |
                  |/ DHCPREQUEST  |  DHCPREQUEST\ |
                  |               |               |
                  |               |     Commits configuration
                  |               |               |
                  |               | _____________/|
                  |               |/ DHCPACK      |
                  |               |               |
                  |    Initialization complete    |
                  |               |               |
                  .               .               .
                  .               .               .
                  |               |               |
                  |      Graceful shutdown        |
                  |               |               |
                  |               |\ ____________ |
                  |               | DHCPRELEASE  \|
                  |               |               |
                  |               |        Discards lease
                  |               |               |
                  v               v               v

 */

// Generate the xid's and send out a discover packet
void dhcp_init(void) {
   dhcp_xid1 = 0x5F31;
   dhcp_xid2 = 0xa929; // make this random!!
   dhcp_send_discover(dhcp_xid1, dhcp_xid2);
}


void dhcp_fillpacket(void) {
  //Set up dhcp packet
  DHCPBUF->htype = 1;                //1 = ethernet
  DHCPBUF->hlen = 6;                 //MAC address is 6 bytes
  DHCPBUF->hops = 0;                 //client always sets to 0
  DHCPBUF->xid[0] = dhcp_xid1;
  DHCPBUF->xid[1] = dhcp_xid2;
  DHCPBUF->flags = 0;                //broadcast bit = 0, all other bits are reserved
  DHCPBUF->ciaddr[0] = 0;               //we have no address yet
  DHCPBUF->ciaddr[1] = 0;
  DHCPBUF->giaddr[0] = 0;
  DHCPBUF->giaddr[1] = 0;
  DHCPBUF->chaddr[0] = UIP_ETHADDR0;
  DHCPBUF->chaddr[1] = UIP_ETHADDR1;
  DHCPBUF->chaddr[2] = UIP_ETHADDR2;
  DHCPBUF->chaddr[3] = UIP_ETHADDR3;
  DHCPBUF->chaddr[4] = UIP_ETHADDR4;
  DHCPBUF->chaddr[5] = UIP_ETHADDR5;
  DHCPBUF->sname[0] = 0;
  DHCPBUF->file[0] = 0;
  DHCPBUF->options[0] = (DHCP_OPTION_COOKIE>>24)&0xFF;
  DHCPBUF->options[1] = (DHCP_OPTION_COOKIE>>16)&0xFF;
  DHCPBUF->options[2] = (DHCP_OPTION_COOKIE>>8)&0xFF;
  DHCPBUF->options[3] = DHCP_OPTION_COOKIE&0xFF;
}

dhcpErrorCode dhcp_send_discover(uint16_t xid1, uint16_t xid2) {
  char namelen, vendorlen;
  namelen= strlen(hostname);
  vendorlen = strlen(vendorname);

  //Sends a DHCPDISCOVER packet

  DHCPBUF->op = DHCP_BOOTREQUEST;    //request
  dhcp_fillpacket();   // with the basic options

  DHCPBUF->secs = 0;                 //starting now

  DHCPBUF->yiaddr[0] = 0;
  DHCPBUF->yiaddr[1] = 0;
  DHCPBUF->siaddr[0] = 0;
  DHCPBUF->siaddr[1] = 0;

  DHCPBUF->options[4] = DHCP_OPT_MESSAGE_TYPE;   //option type
  DHCPBUF->options[5] = 1;                       //option lengths
  DHCPBUF->options[6] = DHCPDISCOVER;            //option
  DHCPBUF->options[7] = 0xfb;
  DHCPBUF->options[8] = 1;
  DHCPBUF->options[9] = 1;
  DHCPBUF->options[10] = DHCP_OPT_CLIENTID;
  DHCPBUF->options[11] = 0x07;
  DHCPBUF->options[12] = 0x01;
  DHCPBUF->options[13] = UIP_ETHADDR0;
  DHCPBUF->options[14] = UIP_ETHADDR1;
  DHCPBUF->options[15] = UIP_ETHADDR2;
  DHCPBUF->options[16] = UIP_ETHADDR3;
  DHCPBUF->options[17] = UIP_ETHADDR4;
  DHCPBUF->options[18] = UIP_ETHADDR5;
  DHCPBUF->options[19] = DHCP_OPT_REQIP;
  DHCPBUF->options[20] = 4;
  DHCPBUF->options[21] = uip_hostaddr[1];
  DHCPBUF->options[22] = uip_hostaddr[1] >> 8;
  DHCPBUF->options[23] = uip_hostaddr[0];
  DHCPBUF->options[24] = uip_hostaddr[0] >> 8;
  DHCPBUF->options[25] = DHCP_OPT_HOSTNAME;
  DHCPBUF->options[26] = namelen;
  memcpy(&(DHCPBUF->options[27]), hostname, namelen);
  DHCPBUF->options[27+namelen] = DHCP_OPT_VENDORID;
  DHCPBUF->options[28+namelen] = vendorlen;
  memcpy(&(DHCPBUF->options[29+namelen]), vendorname, vendorlen);
  DHCPBUF->options[29+vendorlen+namelen] = DHCP_OPT_PARAMREQ;
  DHCPBUF->options[30+vendorlen+namelen] = 6;
  DHCPBUF->options[31+vendorlen+namelen] = DHCP_PARAMREQ_SUBNETMASK;
  DHCPBUF->options[32+vendorlen+namelen] = DHCP_PARAMREQ_DOMAINNAME;
  DHCPBUF->options[33+vendorlen+namelen] = DHCP_PARAMREQ_ROUTER;
  DHCPBUF->options[34+vendorlen+namelen] = DHCP_PARAMREQ_DNS;
  DHCPBUF->options[35+vendorlen+namelen] = DHCP_PARAMREQ_ROUTERDISC;
  DHCPBUF->options[36+vendorlen+namelen] = DHCP_PARAMREQ_STATICROUTE;
  DHCPBUF->options[37+vendorlen+namelen] = DHCP_ENDOPT;
  memset(&(DHCPBUF[38+vendorlen+namelen]), 0x00, 20);
  dhcp_status = DHCP_DISCOVER;
  dhcp_timeout = 10;
  putstring("Sending DHCP discover...");
  dhcp_send_packet();
  return DHCP_SUCCESS;
}


void dhcp_send_request(void) {
  uint8_t namelen;

  DHCPBUF->op = DHCP_BOOTREQUEST;    //request
  dhcp_fillpacket();   // with the basic options
 
  DHCPBUF->options[4] = DHCP_OPT_MESSAGE_TYPE;   //option type
  DHCPBUF->options[5] = 1;                       //option lengths
  DHCPBUF->options[6] = DHCPREQUEST;             //option

  DHCPBUF->options[7] = DHCP_OPT_REQUESTED_IP;   //option type
  DHCPBUF->options[8] = 4;                       //option length
  DHCPBUF->options[9] = uip_hostaddr[0];
  DHCPBUF->options[10] = uip_hostaddr[0]>>8;
  DHCPBUF->options[11] = uip_hostaddr[1];
  DHCPBUF->options[12] = uip_hostaddr[1]>>8;

  DHCPBUF->options[13] = DHCP_OPT_HOSTNAME;
  namelen= strlen(hostname);
  DHCPBUF->options[14] = namelen;
  memcpy(&(DHCPBUF->options[15]), hostname, namelen);
  
  DHCPBUF->options[15+namelen] = 0xFF;

  // terminator
  //printf("send request ");
  dhcp_send_packet();
}

void dhcp_send_packet(void) {
  udp_send_packet(uip_hostaddr, broadcastaddr, 
		  DHCP_CLIENT_PORT, DHCP_SERVER_PORT,
		  300);
  //DHCP_BASE_LEN + 8);  // application data length
}

void dhcp_listen(void) {
  unsigned char timeout;
  int packetlen = 0;
  timeout = getTimer0();
  while (getTimer0() - timeout < 3) {       // ~3 second timeout
    do {
      packetlen = nic_poll();
    } while (packetlen == 0);

    // find a TCP UDP packet
    if ( (ETHBUF->type == htons(UIP_ETHTYPE_IP)) &&
	 (UDPBUF->vhl == 0x45) &&
	 (UDPBUF->proto == UIP_PROTO_UDP) &&
	 (UDPBUF->srcport == HTONS(DHCP_SERVER_PORT)) && 
	 (UDPBUF->destport == HTONS(DHCP_CLIENT_PORT)))
      {
	if ((dhcp_messagetype() == DHCPOFFER) &&
	    (DHCPBUF->xid[0] == dhcp_xid1) &&
	    (DHCPBUF->xid[1] == dhcp_xid2)) {
	  putstring("Received DHCP offer\n\r");
	  //printf("offer (len %d)\n\r", packetlen);
	  dhcp_recv_offer();
	}
	
	return;
      }
  }
}						   

void dhcp_recv_offer() {
  byte *optloc;

  optloc = dhcp_findoption(DHCP_OPT_ROUTER);
  if (optloc != 0) {
    uip_arp_draddr[0] = (optloc[2] << 8) | optloc[1];
    uip_arp_draddr[1] = (optloc[4] << 8) | optloc[3];
    //arpTableEnter(ip);
  } else {
    //printf("couldnt find router option!\n\r");
  }
  
  optloc = dhcp_findoption(DHCP_OPT_NETMASK);
  if (optloc != 0) {
    uip_arp_netmask[0] = (optloc[2] << 8) | optloc[1];
    uip_arp_netmask[1] = (optloc[4] << 8) | optloc[3];
  } else {
    //printf("couldnt find netmask option!\n\r");
  }

  
  optloc = dhcp_findoption(DHCP_OPT_DNS);
  if (optloc != 0) {
    dns_nameserver[0] = (optloc[2] << 8) | optloc[1];
    dns_nameserver[1] = (optloc[4] << 8) | optloc[3];
  } else {
    //printf("couldnt find dns option!\n\r");
  }

   optloc = dhcp_findoption(DHCP_OPT_T1_TIME);
   if (optloc != 0) {
      //dhcp_lease_timeout = optloc[1]; dhcp_lease_timeout <<= 8;
      //dhcp_lease_timeout |= optloc[2]; dhcp_lease_timeout <<= 8;
      //dhcp_lease_timeout |= optloc[3]; dhcp_lease_timeout <<= 8;
      //dhcp_lease_timeout |= optloc[4];
      //printf(", lease timeout in %lx s, ", dhcp_lease_timeout);
   }
   
   uip_hostaddr[0] = DHCPBUF->yiaddr[0]; 
   uip_hostaddr[1] = DHCPBUF->yiaddr[1]; 

   /*
  printf("ip: %4x%4x router: %4x%4x netmask: %4x%4x dns: %4.x%4.x\n\r", 
	 uip_hostaddr[0], uip_hostaddr[1],
	 uip_arp_draddr[0], uip_arp_draddr[1], 
	 uip_arp_netmask[0], uip_arp_netmask[1],
	 dns_nameserver[0], dns_nameserver[1]);
   */
  dhcp_send_request();
}

char dhcp_messagetype() {
  byte i;
   for(i = 4; i <= 128 && DHCPBUF->options[i] != 0xFF; ){
      if(DHCPBUF->options[i] == DHCP_OPT_MESSAGE_TYPE)
         return DHCPBUF->options[i + 2];
      else
         i += DHCPBUF->options[i + 1] + 2;
   }
   printf("couldnt find message type!");
   return 0;
}

uint8_t *dhcp_findoption(uint8_t optionname) {
  int i;
  for (i=4; (i <= DHCP_OPTIONSIZE) && (DHCPBUF->options[i] != 0xff); ) {
    if (DHCPBUF->options[i] == optionname) 
      return (byte *)(DHCPBUF->options + i + 1);
    else
      i += DHCPBUF->options[i+1]+2;
  }
  return 0;
}
