/*****************************************************************************
*  Module Name:       uIP-AVR Port - main control loop shell
*  
*  Created By:        Louis Beaudoin (www.embedded-creations.com)
*
*  Original Release:  September 21, 2002 
*
*  Module Description:  
*  This main control loop shell provides everything required for a basic uIP
*  application using the RTL8019AS NIC
*
*   
*  November 16, 2003
*    Changed NIC interface from RTL8019 specific to general NIC calls
*    Calls the uip_arp_timer function every 10 seconds
*    
*  September 30, 2002
*    Added support for Imagecraft Compiler
*****************************************************************************/
#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/signal.h>
#include <stdio.h>
#include <ctype.h>

#include "uip.h"
#include "nic.h"
#include "uip_arp.h"
#include "main.h"
#include "delay.h"
#include "dhcp.h"
#include "udp.h"
#include "dnslookup.h"
#include "compiler.h"
#include "tcp.h"

uint8_t DEBUG = 0;
uint16_t rand;



void step(void) {
  uart_getchar();
}
/*****************************************************************************
*  Periodic Timout Functions and variables
*
*  The periodic timeout rate can be changed depeding on your application
*  Modify these functions and variables based on your AVR device and clock
*    rate
*  The current setup will interrupt every 256 timer ticks when the timer
*    counter overflows.  timerCounter must count until 0.5 seconds have
*    alapsed
*****************************************************************************/

#define TIMER_PRESCALE    1024
#define BAUDRATE          19200UL
#define TIMERCOUNTER_PERIODIC_TIMEOUT (F_CPU / TIMER_PRESCALE / 20)

static unsigned char timerCounter, tick;

unsigned char getTimer0(void) {
  return tick;
}
unsigned char getTimerCounter() {
  return timerCounter;
}

void initTimer(void)
{
  outp( 7, TCCR0 ) ;  // timer0 prescale 1/1024 (7)

  // interrupt on overflow
  sbi( TIMSK, TOIE0 ) ;
	
  timerCounter = 0;
}


SIGNAL(SIG_OVERFLOW0)
{
  if (timerCounter++ % 64 == 0) {
    tick++;                // ~ 1 second
  }
}


/*************** our code **************************/

uint8_t http_state = 0;
void http_established(tcp_conn *conn, uint8_t *data, uint16_t len) {
  int i;
  if (len != 0) {
    //printf("\n\r##### Recvd %d bytes #######\n\r", len);
  }
  switch (http_state) {
  case 0:  // just established connection
    data = (uint8_t *)TCPBUF + 40;     // 20 bytes for IP header 8 bytes for UDP header
    putstring("Sending request...\n\r");
    sprintf(data, "GET / HTTP/1.0\n\r\n\r");
    tcp_send(conn, data, 19, &http_established); // call back to me when we get resp
    putstring("wait...\n\r");
    http_state = 1;
    break;
    
  case 1:
    //putstring("woo>\n\r");
    for (i=0; i<len; i++)
      uart_putchar(data[i]);
    
    break;
  }
}

void send_periodic_udp(void) // send some data, periodically! 
{
  byte *data;
  int len;
  uint16_t destaddr[2] = {0x5512, 0xee01}; //18.85.1.238
  uint16_t srcport = 4321;
  uint16_t destport = 1234;
  
  data = (byte *)UDPBUF + 28;     // 20 bytes for IP header 8 bytes for UDP header
  printf("sending periodic packet\n\r");
  sprintf(data, "hello!");
  len = 6;
  udp_send_packet(uip_hostaddr, &destaddr, // send from me to them
		  srcport, destport,
		  len);
}

void handle_udp_port_9999(char *data, int len) {
  int i;
  uint16_t destaddr[2];
  
  printf("* ");
  // umm, do something with the data like print it out
  for (i=0; i<len; i++)
    uart_putchar(data[i]);
  printf(" *\n\r");
  
  // now send the data back, but upper cased!

  destaddr[0] = UDPBUF->srcipaddr[0];
  destaddr[1] = UDPBUF->srcipaddr[1];
  printf("sending data back to %d.%d.%d.%d ", 
	 destaddr[0] & 0xFF, destaddr[0] >> 8, 
	 destaddr[1] & 0xFF, destaddr[1] >> 8);
  
  for (i=0; i<len; i++)
    data[i] = toupper(data[i]);

  udp_send_packet(uip_hostaddr, &destaddr, // send from me to them
		  NTOHS(UDPBUF->destport), // swap dest and src prot
		  NTOHS(UDPBUF->srcport),
		  len);

}
  

/*****************************************************************************
*  Main Control Loop
*
*  
*****************************************************************************/
int main(void)
{
  unsigned char arptimer=0;
  int i;
  uint16_t tmp16, addr[2];
  
  // setup IO's and UART 
  ioinit();


  putstring("uIP/Aether for cs8900a build 10/1/2004\n\r");


  // init NIC device driver
  nic_init();


  // init ARP cache
  uip_arp_init();

  // init periodic timer
  initTimer();
  
  sei();
  
#if UIP_FIXEDADDR == 0
  if (uip_hostaddr[0] == 0 && uip_hostaddr[1] == 0) {
    dhcp_init();
    dhcp_listen();
  }
#endif

  putstring("My addr: ");
  printIP(uip_hostaddr);
  putstring(" router: ");
  printIP(uip_arp_draddr);
  putstring(" netmask: ");
  printIP(uip_arp_netmask);
  putstring(" dns: ");
  printIP(dns_nameserver);
  putstring("\n\r");

  rand = timerCounter; rand <<= 8;
  rand |= timerCounter;

  tcp_init();

  if (! (dns_nameserver[0] == 0 && dns_nameserver[1] == 0) ) {
    i = lookup("www.google.com"); // lookup is a blocking function
    if (i == DNS_NOSUCHNAME) {
      putstring("failed lookup\n\r");
    } else if (i == DNS_TIMEOUT) {
      putstring("timed out\n\r");
    } else {
      putstring("google is ");
      printIP(dns_resolved);
      putstring("\n\r");
      // when a function is passed, tcp_open is nonblocking,
      // ie it returns immediately, and the function is called
      // when the connection is established
      tcp_open(dns_resolved, 80, &http_established);
    }
  }
 



  while(1)
    {
      // look for a packet
      uip_len = nic_poll();
      if(uip_len == 0)
	{
	  // if timed out, call periodicfunction for each connection
	  if(timerCounter > 100) // ~1second
	    {
	      timerCounter = 0;

	      //send_periodic_udp();
	      /* Call the ARP timer function every 10 seconds. */
	      if(++arptimer == 10)
		{	
#if UIP_FIXEDADDR == 0
		  if (uip_hostaddr[0] == 0 && uip_hostaddr[1] == 0) {
		    dhcp_init();
		    dhcp_listen();
		  }
#endif
		  uip_arp_timer();
		  arptimer = 0;
		}
	    }
	}
      else  // packet received
	{
	  // process an IP packet
	  if((ETHBUF->type == htons(UIP_ETHTYPE_IP)) &&
	     (IPBUF->vhl == 0x45))
	    {
	      // add the source to the ARP cache
	      // also correctly set the ethernet packet length before processing
	      uip_arp_ipin();
	      
	      //uip_input();
	      //printf("IP for ");
	      //printIP(IPBUF->destipaddr);
	      //printf(" from ");
	      //printIP(IPBUF->srcipaddr);

	      // determine what kind of packet it is

	      // ICMP? (ping)
	      if ((IPBUF->proto == UIP_PROTO_ICMP) &&
		  (ICMPBUF->type == ICMP_ECHO)) {
		//printf(" ICMP echo request...");

		// reply to it
		ICMPBUF->type = ICMP_ECHO_REPLY;
  
		// fix the checksum
		if(ICMPBUF->icmpchksum >= HTONS(0xffff - (ICMP_ECHO << 8))) {
		  ICMPBUF->icmpchksum += HTONS(ICMP_ECHO << 8) + 1;
		} else {
		  ICMPBUF->icmpchksum += HTONS(ICMP_ECHO << 8);
		}
  
		/* Swap IP addresses. */
		tmp16 = IPBUF->destipaddr[0];
		IPBUF->destipaddr[0] = IPBUF->srcipaddr[0];
		IPBUF->srcipaddr[0] = tmp16;
		tmp16 = IPBUF->destipaddr[1];
		IPBUF->destipaddr[1] = IPBUF->srcipaddr[1];
		IPBUF->srcipaddr[1] = tmp16;

		//printf("getting arp...");
		uip_arp_out();       // fill in the right arp addr
		
		NICBeginPacketSend(uip_len);
		
		NICSendPacketData(uip_buf, uip_len);
		NICEndPacketSend();
		//printf(" replied\n\r");
	      }

	      else if (IPBUF->proto == UIP_PROTO_UDP) {
		/*printf (" UDP packet (%u -> %u) %u bytes\n\r",
			NTOHS(UDPBUF->srcport),
			NTOHS(UDPBUF->destport),
			NTOHS(UDPBUF->udplen) - 8);
		*/
		// call our listening function(s)
		switch (NTOHS(UDPBUF->destport)) 
		  {
		  case 9999:
		    handle_udp_port_9999((byte *)UDPBUF+28,
					 NTOHS(UDPBUF->udplen) - 8);
		    // the udpip header is 28 bytes long so index past that
		    // and the length includes the header so subract that too
		    break;
		  default:
		    break;   // dont do anything for any other ports
		  }
	      }
	      else if (IPBUF->proto == UIP_PROTO_TCP) {
		tcp_process();
	      }
	      else
		{
		  //printf(" unknown type\n\r");
		}
	    }
	  // process an ARP packet
	  else if(ETHBUF->type == htons(UIP_ETHTYPE_ARP))
	    {
	      uip_arp_arpin();
	      //uart_putchar('a');
	      // transmit a packet, if one is ready
	      if(uip_len > 0)
		nic_send();
	    }
	}
    }
  return 1;
}


void printIP(uint16_t *addr) {
  printf("%d.%d.%d.%d", 
	 addr[0] & 0xFF, addr[0] >> 8,
	 addr[1] & 0xFF, addr[1] >> 8);
}
	      
// setup IO and UART
int uart_putchar(char c)
{
   loop_until_bit_is_set(UCSRA, UDRE);
   UDR = c;
   return 0;
}


int uart_getchar(void) {
  char c;
  loop_until_bit_is_set(UCSRA, RXC);
  c = UDR;
  return (int)c;
}

void ioinit() {
  uint16_t baud = (F_CPU / (16 * BAUDRATE)) - 1;
  
  /* setup the main UART */
  UCSRB |= (1<<RXEN) | (1<<TXEN);    // read and write
  UBRRL = (uint8_t)baud;               // set baudrate
  UBRRH = (uint8_t)(baud>>8);

  /* attach the UART to stdio */
  fdevopen(NULL, uart_getchar, 0);   // opens stdin to uart
  fdevopen(uart_putchar, NULL, 0);   // opens stdout to uart
  outb(DDRA, 0xFF);
  outb(DDRB, 0xFF);
  outb(DDRC, 0xFF);
  outb(DDRD, 0xFF);

  
}

void putstring(char *str) {
  while (str[0] != 0) {
    uart_putchar(str[0]);
    str++;
  }
}

///
void halt() {
  printf("halting");
  // turn off interrupts??
  while (1) {
  }
}


uint16_t rnd() {
  rand = ((((rand >> 15) ^ (rand >> 14) ^ (rand >> 12) ^ (rand >> 3))
	       & 0000000000000001)
	      | (rand << 1)); /*Or with the register shifted right.*/
  return rand; /*Return the first bit.*/
}
