/**
 * \addtogroup uip
 * @{
 */

/**
 * \defgroup uiparp uIP Address Resolution Protocol
 * @{
 * 
 * The Address Resolution Protocol ARP is used for mapping between IP
 * addresses and link level addresses such as the Ethernet MAC
 * addresses. ARP uses broadcast queries to ask for the link level
 * address of a known IP address and the host which is configured with
 * the IP address for which the query was meant, will respond with its
 * link level address.
 *
 * \note This ARP implementation only supports Ethernet.
 */
 
/**
 * \file
 * Implementation of the ARP Address Resolution Protocol.
 * \author Adam Dunkels <adam@dunkels.com>
 *
 */

/*
 * Copyright (c) 2001-2003, Adam Dunkels.
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met: 
 * 1. Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright 
 *    notice, this list of conditions and the following disclaimer in the 
 *    documentation and/or other materials provided with the distribution. 
 * 3. The name of the author may not be used to endorse or promote
 *    products derived from this software without specific prior
 *    written permission.  
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.  
 *
 * This file is part of the uIP TCP/IP stack.
 *
 * $Id: uip_arp.c,v 1.1 2004/10/06 20:17:17 ladyada Exp $
 *
 */


#include "uip_arp.h"
#include "nic.h"
#include "main.h"
#include <string.h>
#include <stdio.h>

struct arp_hdr {
  struct uip_eth_hdr ethhdr;
  uint16_t hwtype;
  uint16_t protocol;
  u8_t hwlen;
  u8_t protolen;
  uint16_t opcode;
  struct uip_eth_addr shwaddr;
  uint16_t sipaddr[2];
  struct uip_eth_addr dhwaddr;
  uint16_t dipaddr[2]; 
};

struct ethip_hdr {
  struct uip_eth_hdr ethhdr;
  /* IP header. */
  u8_t vhl,
    tos,          
    len[2],       
    ipid[2],        
    ipoffset[2],  
    ttl,          
    proto;     
  uint16_t ipchksum;
  uint16_t srcipaddr[2], 
    destipaddr[2];
};

#define ARP_REQUEST 1
#define ARP_REPLY   2

#define ARP_HWTYPE_ETH 1

struct arp_entry {
  uint16_t ipaddr[2];
  struct uip_eth_addr ethaddr;
  u8_t time;
};

struct uip_eth_addr uip_ethaddr = {{UIP_ETHADDR0,
				    UIP_ETHADDR1,
				    UIP_ETHADDR2,
				    UIP_ETHADDR3,
				    UIP_ETHADDR4,
				    UIP_ETHADDR5}};

static struct arp_entry arp_table[UIP_ARPTAB_SIZE];
static uint16_t ipaddr[2];
static u8_t i, c;

static u8_t arptime;
static u8_t tmpage;

#define ARPBUF   ((struct arp_hdr *)&uip_buf[0])
#define ETHIPBUF ((struct ethip_hdr *)&uip_buf[0])
/*-----------------------------------------------------------------------------------*/
/**
 * Initialize the ARP module.
 *
 */
/*-----------------------------------------------------------------------------------*/
void
uip_arp_init(void)
{
  for(i = 1; i < UIP_ARPTAB_SIZE; ++i) {
    memset(arp_table[i].ipaddr, 0, 4);
  }

  // first is always broadcast!
  memset(arp_table[0].ipaddr, 0xff, 4);
  memset(arp_table[0].ethaddr.addr, 0xff, 6);
  arp_table[0].time = 0;
}
/*-----------------------------------------------------------------------------------*/
/**
 * Periodic ARP processing function.
 *
 * This function performs periodic timer processing in the ARP module
 * and should be called at regular intervals. The recommended interval
 * is 10 seconds between the calls.
 *
 */
/*-----------------------------------------------------------------------------------*/
void
uip_arp_timer(void)
{
  struct arp_entry *tabptr;
  
  ++arptime;
  // ignore the first entry, it is the broadcast arp
  for(i = 1; i < UIP_ARPTAB_SIZE; ++i) {
    tabptr = &arp_table[i];
    if((tabptr->ipaddr[0] | tabptr->ipaddr[1]) != 0 &&
       arptime - tabptr->time >= UIP_ARP_MAXAGE) {
      memset(tabptr->ipaddr, 0, 4);
    }
  }

}
/*-----------------------------------------------------------------------------------*/
static char 
uip_arp_update(uint16_t *ipaddr, struct uip_eth_addr *ethaddr)
{
  register struct arp_entry *tabptr;
  /* Walk through the ARP mapping table and try to find an entry to
     update. If none is found, the IP -> MAC address mapping is
     inserted in the ARP table.
     Skip the first entry, its broadcast and always stays */
  for(i = 1; i < UIP_ARPTAB_SIZE; ++i) {

    tabptr = &arp_table[i];
    /* Only check those entries that are actually in use. */
    if(tabptr->ipaddr[0] != 0 &&
       tabptr->ipaddr[1] != 0) {

      /* Check if the source IP address of the incoming packet matches
         the IP address in this ARP table entry. */
      if(ipaddr[0] == tabptr->ipaddr[0] &&
	 ipaddr[1] == tabptr->ipaddr[1]) {
	
	/* An old entry found, update this and return. */
	memcpy(tabptr->ethaddr.addr, ethaddr->addr, 6);
	tabptr->time = arptime;

	if (DEBUG) printf("placed arp data in entry #%d\n\r", i);
	return i;
      }
    }
  }

  /* If we get here, no existing ARP table entry was found, so we
     create one. */

  /* First, we try to find an unused entry in the ARP table. */
  for(i = 1; i < UIP_ARPTAB_SIZE; ++i) {
    tabptr = &arp_table[i];
    if(tabptr->ipaddr[0] == 0 &&
       tabptr->ipaddr[1] == 0) {
      break;
    }
  }

  /* If no unused entry is found, we try to find the oldest entry and
     throw it away. */
  if(i == UIP_ARPTAB_SIZE) {
    tmpage = 0;
    c = 0;
    for(i = 1; i < UIP_ARPTAB_SIZE; ++i) {
      tabptr = &arp_table[i];
      if(arptime - tabptr->time > tmpage) {
	tmpage = arptime - tabptr->time;
	c = i;
      }
    }
    i = c;
  }
  
  if (DEBUG) printf("inserted @ %d\n\r", i);
  /* Now, i is the ARP table entry which we will fill with the new
     information. */
  memcpy(tabptr->ipaddr, ipaddr, 4);
  memcpy(tabptr->ethaddr.addr, ethaddr->addr, 6);
  tabptr->time = arptime;
  return i;
}
/*-----------------------------------------------------------------------------------*/
/**
 * ARP processing for incoming IP packets
 *
 * This function should be called by the device driver when an IP
 * packet has been received. The function will check if the address is
 * in the ARP cache, and if so the ARP cache entry will be
 * refreshed. If no ARP cache entry was found, a new one is created.
 *
 * This function expects an IP packet with a prepended Ethernet header
 * in the uip_buf[] buffer, and the length of the packet in the global
 * variable uip_len.
 */
/*-----------------------------------------------------------------------------------*/
void
uip_arp_ipin(void)
{
  uip_len -= sizeof(struct uip_eth_hdr);
	
  /* Only insert/update an entry if the source IP address of the
     incoming IP packet comes from a host on the local network. */
  if((ETHIPBUF->srcipaddr[0] & uip_arp_netmask[0]) !=
     (uip_hostaddr[0] & uip_arp_netmask[0])) {
    return;
  }
  if((ETHIPBUF->srcipaddr[1] & uip_arp_netmask[1]) !=
     (uip_hostaddr[1] & uip_arp_netmask[1])) {
    return;
  }
  uip_arp_update(ETHIPBUF->srcipaddr, &(ETHIPBUF->ethhdr.src));
  
  return;
}
/*-----------------------------------------------------------------------------------*/
/**
 * ARP processing for incoming ARP packets.
 *
 * This function should be called by the device driver when an ARP
 * packet has been received. The function will act differently
 * depending on the ARP packet type: if it is a reply for a request
 * that we previously sent out, the ARP cache will be filled in with
 * the values from the ARP reply. If the incoming ARP packet is an ARP
 * request for our IP address, an ARP reply packet is created and put
 * into the uip_buf[] buffer.
 *
 * When the function returns, the value of the global variable uip_len
 * indicates whether the device driver should send out a packet or
 * not. If uip_len is zero, no packet should be sent. If uip_len is
 * non-zero, it contains the length of the outbound packet that is
 * present in the uip_buf[] buffer.
 *
 * This function expects an ARP packet with a prepended Ethernet
 * header in the uip_buf[] buffer, and the length of the packet in the
 * global variable uip_len.
 */
/*-----------------------------------------------------------------------------------*/
void
uip_arp_arpin(void)
{

  if(uip_len < sizeof(struct arp_hdr)) {
    uip_len = 0;
    return;
  }

  uip_len = 0;
  
  switch(ARPBUF->opcode) {
  case HTONS(ARP_REQUEST):
    /* ARP request. If it asked for our address, we send out a
       reply. */
    if(ARPBUF->dipaddr[0] == uip_hostaddr[0] &&
       ARPBUF->dipaddr[1] == uip_hostaddr[1]) {
      /* The reply opcode is 2. */
      ARPBUF->opcode = HTONS(2);

      memcpy(ARPBUF->dhwaddr.addr, ARPBUF->shwaddr.addr, 6);
      memcpy(ARPBUF->shwaddr.addr, uip_ethaddr.addr, 6);
      memcpy(ARPBUF->ethhdr.src.addr, uip_ethaddr.addr, 6);
      memcpy(ARPBUF->ethhdr.dest.addr, ARPBUF->dhwaddr.addr, 6);
      
      ARPBUF->dipaddr[0] = ARPBUF->sipaddr[0];
      ARPBUF->dipaddr[1] = ARPBUF->sipaddr[1];
      ARPBUF->sipaddr[0] = uip_hostaddr[0];
      ARPBUF->sipaddr[1] = uip_hostaddr[1];

      ARPBUF->ethhdr.type = HTONS(UIP_ETHTYPE_ARP);      
      uip_len = sizeof(struct arp_hdr);
    }      
    break;
  case HTONS(ARP_REPLY):
    /* ARP reply. We insert or update the ARP table if it was meant
       for us. */
    if(ARPBUF->dipaddr[0] == uip_hostaddr[0] &&
       ARPBUF->dipaddr[1] == uip_hostaddr[1]) {

      uip_arp_update(ARPBUF->sipaddr, &ARPBUF->shwaddr);
    }
    break;
  }

  return;
}
/*-----------------------------------------------------------------------------------*/
/**
 * Prepend Ethernet header to an outbound IP packet and see if we need
 * to send out an ARP request.
 *
 * This function should be called before sending out an IP packet. The
 * function checks the destination IP address of the IP packet to see
 * what Ethernet MAC address that should be used as a destination MAC
 * address on the Ethernet.
 *
 * If the destination IP address is in the local network (determined
 * by logical ANDing of netmask and our IP address), the function
 * checks the ARP cache to see if an entry for the destination IP
 * address is found. If so, an Ethernet header is prepended and the
 * function returns. If no ARP cache entry is found for the
 * destination IP address, the packet in the uip_buf[] is replaced by
 * an ARP request packet for the IP address. The IP packet is dropped
 * and it is assumed that they higher level protocols (e.g., TCP)
 * eventually will retransmit the dropped packet.
 *
 * If the destination IP address is not on the local network, the IP
 * address of the default router is used instead.
 *
 * When the function returns, a packet is present in the uip_buf[]
 * buffer, and the length of the packet is in the global variable
 * uip_len.
 */
/*-----------------------------------------------------------------------------------*/
void
uip_arp_out(void)
{
  struct arp_entry *tabptr;
  /* Find the destination IP address in the ARP table and construct
     the Ethernet header. If the destination IP addres isn't on the
     local network, we use the default router's IP address instead.

     If not ARP table entry is found, we overwrite the original IP
     packet with an ARP request for the IP address. */

  /* Check if the destination address is on the local network. */
  if (ETHIPBUF->destipaddr[0] == 0xffff && ETHIPBUF->destipaddr[1] == 0xffff) {
    ipaddr[0] = 0xffff;
    ipaddr[1] = 0xffff;
  }
  else if ((ETHIPBUF->destipaddr[0] & uip_arp_netmask[0]) !=
     (uip_hostaddr[0] & uip_arp_netmask[0]) ||
     (ETHIPBUF->destipaddr[1] & uip_arp_netmask[1]) !=
     (uip_hostaddr[1] & uip_arp_netmask[1])) {
    /* Destination address was not on the local network, so we need to
       use the default router's IP address instead of the destination
       address when determining the MAC address. */
    if (DEBUG) printf("not local ip!");
    ipaddr[0] = uip_arp_draddr[0];
    ipaddr[1] = uip_arp_draddr[1];
  } else {
    /* Else, we use the destination IP address. */
    ipaddr[0] = ETHIPBUF->destipaddr[0];
    ipaddr[1] = ETHIPBUF->destipaddr[1];
  }
  

  for(i = 0; i < UIP_ARPTAB_SIZE + 1; i++) {


    tabptr = &arp_table[i];
    //printf("%d: %x%x == %x%x? ", i, ipaddr[0], ipaddr[1], 
    //   tabptr->ipaddr[0],  tabptr->ipaddr[1]);
    if(ipaddr[0] == tabptr->ipaddr[0] &&
       ipaddr[1] == tabptr->ipaddr[1]) {
      if (DEBUG) printf("found the arp entry\n\r");
      break;
    } else {
      if (DEBUG) printf("didnt find arp entry\n\r");
    }
  }

  if(i == UIP_ARPTAB_SIZE) {
    // at the end of the table! quick! send out an arp!
    if (DEBUG) printf("sending arp for %x%x\n\r", ipaddr[0], ipaddr[1]);
    i = uip_arp_getarp(ipaddr);
    if (i == 0) 
      return;
    tabptr = &arp_table[i];
  }

  /* Build an ethernet header. */
  memcpy(ETHIPBUF->ethhdr.dest.addr, tabptr->ethaddr.addr, 6);
  memcpy(ETHIPBUF->ethhdr.src.addr, uip_ethaddr.addr, 6);
  
  ETHIPBUF->ethhdr.type = HTONS(UIP_ETHTYPE_IP);

  uip_len += sizeof(struct uip_eth_hdr);
}



struct arp_hdr arpbuf;
char uip_arp_getreply(uint16_t *ipaddr) {
  unsigned char timeout;
  int packetlen = 0;
   
  timeout = getTimer0();
  while (getTimer0() - timeout < 3) {       // ~3 second timeout
    while (packetlen == 0) {
      packetlen = NICBeginPacketRetreive();
    }
    NICRetreivePacketData((uchar *)&arpbuf, sizeof(struct arp_hdr));
    NICEndPacketRetreive();
    if ((arpbuf.ethhdr.type == HTONS(UIP_ETHTYPE_ARP)) &&
	(arpbuf.opcode == HTONS(ARP_REPLY)) &&
	(arpbuf.hwtype == HTONS(ARP_HWTYPE_ETH)) &&
	(arpbuf.protocol == HTONS(UIP_ETHTYPE_IP)) &&
	(arpbuf.sipaddr[0] == ipaddr[0]) &&
      	(arpbuf.sipaddr[1] == ipaddr[1])) {
      if (DEBUG) printf("got reply!\n\r");
      // put it in the table!
      return uip_arp_update(&(arpbuf.sipaddr[0]), &(arpbuf.shwaddr));
    }
  }
  if (DEBUG)  printf(" timeout\n\r");
  return 0;
}

// send a few arps and wait for replies
char uip_arp_getarp(uint16_t *ipaddr) {
  char i, tabloc;
  for (i=0; i<5; i++) {
    //printf("arp out\n\r");
    uip_arp_sendreq(ipaddr);
    tabloc = uip_arp_getreply(ipaddr);
    if (tabloc != 0)
      return tabloc;
  }
  return 0;
}

void uip_arp_sendreq(uint16_t *ipaddr) {
  NICBeginPacketSend(sizeof(struct arp_hdr));   // all arps are 42 bytes long
  
  memset(arpbuf.ethhdr.dest.addr, 0xff, 6);
  memset(arpbuf.dhwaddr.addr, 0x00, 6);
  memcpy(arpbuf.ethhdr.src.addr, uip_ethaddr.addr, 6);
  memcpy(arpbuf.shwaddr.addr, uip_ethaddr.addr, 6);
  
  arpbuf.dipaddr[0] = ipaddr[0];
  arpbuf.dipaddr[1] = ipaddr[1];
  arpbuf.sipaddr[0] = uip_hostaddr[0];
  arpbuf.sipaddr[1] = uip_hostaddr[1];
  arpbuf.opcode = HTONS(ARP_REQUEST); /* ARP request. */
  arpbuf.hwtype = HTONS(ARP_HWTYPE_ETH);
  arpbuf.protocol = HTONS(UIP_ETHTYPE_IP);
  arpbuf.hwlen = 6;
  arpbuf.protolen = 4;
  arpbuf.ethhdr.type = HTONS(UIP_ETHTYPE_ARP);
   
  NICSendPacketData((uchar *)&arpbuf, sizeof(struct arp_hdr));
  
  NICEndPacketSend();
}
/*-----------------------------------------------------------------------------------*/

/** @} */
/** @} */
